#!/usr/bin/env python

# Edit this script to add your team's code. Some functions are *required*, but you can edit most parts of the required functions,
# change or remove non-required functions, and add your own functions.

################################################################################
#
# Import libraries and functions. You can change or remove them.
#
################################################################################

from helper_code import *
import numpy as np, scipy as sp, scipy.stats, os, sys, joblib, scipy.signal
from sklearn.impute import KNNImputer, SimpleImputer
from sklearn.ensemble import RandomForestClassifier

#import librosa

from imblearn.under_sampling import ClusterCentroids

sample_frequency = 4000
resample_frequency = 2000
re_resample_frequency = 500

################################################################################
#
# Required functions. Edit these functions to add your code, but do not change the arguments.
#
################################################################################

# Train your model.
def train_challenge_model(data_folder, model_folder, verbose):
    # Find data files.
    if verbose >= 1:
        print('Finding data files...')

    # Find the patient data files.
    patient_files = find_patient_files(data_folder)
    num_patient_files = len(patient_files)

    if num_patient_files==0:
        raise Exception('No data was provided.')

    # Create a folder for the model if it does not already exist.
    os.makedirs(model_folder, exist_ok=True)

    # Extract the features and labels.
    if verbose >= 1:
        print('Extracting features and labels from the Challenge data...')

    murmur_classes = ['Present', 'Unknown', 'Absent']
    num_murmur_classes = len(murmur_classes)
    outcome_classes = ['Abnormal', 'Normal']
    num_outcome_classes = len(outcome_classes)

    features = list()
    murmurs = list()
    outcomes = list()

    for i in range(num_patient_files):
        if verbose >= 2:
            print('    {}/{}...'.format(i+1, num_patient_files))

        # Load the current patient data and recordings.
        current_patient_data = load_patient_data(patient_files[i])
        current_recordings = load_recordings(data_folder, current_patient_data)

        # Extract features.
        current_features = get_features(current_patient_data, current_recordings)
        features.append(current_features)

        # Extract labels and use one-hot encoding.
        current_murmur = np.zeros(num_murmur_classes, dtype=int)
        murmur = get_murmur(current_patient_data)
        if murmur in murmur_classes:
            j = murmur_classes.index(murmur)
            current_murmur[j] = 1
        murmurs.append(current_murmur)

        current_outcome = np.zeros(num_outcome_classes, dtype=int)
        outcome = get_outcome(current_patient_data)
        if outcome in outcome_classes:
            j = outcome_classes.index(outcome)
            current_outcome[j] = 1
        outcomes.append(current_outcome)

    features = np.vstack(features)
    murmurs = np.vstack(murmurs)
    outcomes = np.vstack(outcomes)

    # Train the model.
    if verbose >= 1:
        print('Training model...')

    # Define parameters for random forest classifier.
    n_estimators   = 123  # Number of trees in the forest.
    max_leaf_nodes = 45   # Maximum number of leaf nodes in each tree.
    random_state   = 6789 # Random state; set for reproducibility.

    imputer = SimpleImputer().fit(features)
    features = imputer.transform(features)
    
    #imputer, features = get_data_imputed(features)
    
    # features, labels = get_undersampled(features, labels)
    
    murmur_classifier = RandomForestClassifier(n_estimators=n_estimators, max_leaf_nodes=max_leaf_nodes, random_state=random_state).fit(features, murmurs)
    outcome_classifier = RandomForestClassifier(n_estimators=n_estimators, max_leaf_nodes=max_leaf_nodes, random_state=random_state).fit(features, outcomes)

    # Save the model.
    save_challenge_model(model_folder, imputer, murmur_classes, murmur_classifier, outcome_classes, outcome_classifier)

    if verbose >= 1:
        print('Done.')

# Load your trained model. This function is *required*. You should edit this function to add your code, but do *not* change the
# arguments of this function.
def load_challenge_model(model_folder, verbose):
    filename = os.path.join(model_folder, 'model.sav')
    return joblib.load(filename)

# Run your trained model. This function is *required*. You should edit this function to add your code, but do *not* change the
# arguments of this function.
def run_challenge_model(model, data, recordings, verbose):
    imputer = model['imputer']
    murmur_classes = model['murmur_classes']
    murmur_classifier = model['murmur_classifier']
    outcome_classes = model['outcome_classes']
    outcome_classifier = model['outcome_classifier']

    # Load features.
    features = get_features(data, recordings)

    # Impute missing data.
    features = features.reshape(1, -1)
    features = imputer.transform(features)

    # Get classifier probabilities.
    murmur_probabilities = murmur_classifier.predict_proba(features)
    murmur_probabilities = np.asarray(murmur_probabilities, dtype=np.float32)[:, 0, 1]
    outcome_probabilities = outcome_classifier.predict_proba(features)
    outcome_probabilities = np.asarray(outcome_probabilities, dtype=np.float32)[:, 0, 1]

    # Choose label with highest probability.
    murmur_labels = np.zeros(len(murmur_classes), dtype=np.int_)
    idx = np.argmax(murmur_probabilities)
    murmur_labels[idx] = 1
    outcome_labels = np.zeros(len(outcome_classes), dtype=np.int_)
    idx = np.argmax(outcome_probabilities)
    outcome_labels[idx] = 1

    # Concatenate classes, labels, and probabilities.
    classes = murmur_classes + outcome_classes
    labels = np.concatenate((murmur_labels, outcome_labels))
    probabilities = np.concatenate((murmur_probabilities, outcome_probabilities))

    return classes, labels, probabilities

################################################################################
#
# Optional functions. You can change or remove these functions and/or add new functions.
#
################################################################################

# Save your trained model.
def save_challenge_model(model_folder, imputer, murmur_classes, murmur_classifier, outcome_classes, outcome_classifier):
    d = {'imputer': imputer, 'murmur_classes': murmur_classes, 'murmur_classifier': murmur_classifier, 'outcome_classes': outcome_classes, 'outcome_classifier': outcome_classifier}
    filename = os.path.join(model_folder, 'model.sav')
    joblib.dump(d, filename, protocol=0)

# Extract features from the data.
def get_features(data, recordings):
    # Extract the age group and replace with the (approximate) number of months for the middle of the age group.
    age_group = get_age(data)

    if compare_strings(age_group, 'Neonate'):
        age = 0.5
    elif compare_strings(age_group, 'Infant'):
        age = 6
    elif compare_strings(age_group, 'Child'):
        age = 6 * 12
    elif compare_strings(age_group, 'Adolescent'):
        age = 15 * 12
    elif compare_strings(age_group, 'Young Adult'):
        age = 20 * 12
    else:
        age = float('nan')

    # Extract sex. Use one-hot encoding.
    sex = get_sex(data)

    sex_features = np.zeros(2, dtype=int)
    if compare_strings(sex, 'Female'):
        sex_features[0] = 1
    elif compare_strings(sex, 'Male'):
        sex_features[1] = 1

    # Extract height and weight.
    height = get_height(data)
    weight = get_weight(data)

    # Extract pregnancy status.
    is_pregnant = get_pregnancy_status(data)

    # Extract recording locations and data. Identify when a location is present, and compute the mean, variance, and skewness of
    # each recording. If there are multiple recordings for one location, then extract features from the last recording.
    locations = get_locations(data)

    recording_locations = ['AV', 'MV', 'PV', 'TV', 'Phc']
    num_recording_locations = len(recording_locations)
    recording_features = np.zeros((num_recording_locations, 4+3001), dtype=float) #+4000
    num_locations = len(locations)
    num_recordings = len(recordings)
    
    if num_locations==num_recordings:
        for i in range(num_locations):
            for j in range(num_recording_locations):
                if compare_strings(locations[i], recording_locations[j]) and np.size(recordings[i])>0:
                
                    #recording = scipy.signal.resample(recordings[i], len(recordings[1])/2)
                    recording = recordings[i]
                    # resample_frequency
                    #recording = np.asarray(_normalize(recordings[i]), dtype=np.float32)
                    recording_features[j, 0] = 1
                    recording_features[j, 1] = np.mean(recording)
                    recording_features[j, 2] = np.var(recording)
                    recording_features[j, 3] = sp.stats.skew(recording)
                    
                    f, Pxx_den = scipy.signal.welch(recording, sample_frequency, nperseg=6000, detrend="linear")#, nfft=1024, detrend=True)
                    
                    for k in range(len(Pxx_den)):
                        recording_features[j, k+4] = Pxx_den[k]
                        
                    # mfccs = librosa.feature.mfcc(y=recording, sr=sample_frequency, n_mfcc=21)
                    # mfccs = mfccs.flatten()
                    
                    # for l in range(len(mfccs)):
                        # recording_features[j, k+4+2001] = mfccs[l]
                        
                    # p_init = np.array([0.31, 0.09, 0.6])
                    # p_emit = np.array([[0.69, 0.31],
                                       # [0.50, 0.50],
                                       # [0.20, 0.80]])
                    # p_trans = np.array([[0.8, 0.1, 0.1], [0.6, 0.3, 0.1], [0.3, 0.5, 0.2]])
                    # path, logp = librosa.sequence.viterbi(p_emit, p_trans, p_init=p_init,
                                                          # return_logp=True)
                    # print(logp, path)
                    
                    #librosa.feature.chroma_stft(y=y, sr=sr)
                    
                    #recording = librosa.resample(recording, orig_sr=sample_frequency, target_sr=re_resample_frequency)
                    
                    # chroma_cens = librosa.feature.chroma_cens(y=recording, sr=sample_frequency * 2.5, n_chroma = 5, bins_per_octave = 10)
                    # chroma_cq = librosa.feature.chroma_cqt(y=recording, sr=sample_frequency * 2.5, n_chroma = 5, bins_per_octave = 10)
                    
                    # print(chroma_cens.shape, chroma_cq.shape)
                    
                        
    recording_features = recording_features.flatten()

    features = np.hstack(([age], sex_features, [height], [weight], [is_pregnant], recording_features))

    return np.asarray(features, dtype=np.float32)
    
def get_data_imputed(features):
    demo_features = features[:,0:6]
    
    imputer = KNNImputer(n_neighbors=3, weights='distance').fit(demo_features)
    demo_features = imputer.transform(demo_features)
    
    features = np.concatenate((demo_features[:,0:6], features[:,6:]), axis = 1)
    imputer = KNNImputer(n_neighbors=3, weights='distance').fit(features)
    
    return imputer, features
    

def get_undersampled(features, labels):
    cc = ClusterCentroids(random_state=31)
    features, labels = cc.fit_resample(features, labels)
    
    # sm = SMOTE(sampling_strategy='auto', random_state=37, k_neighbors=5)
    # features, labels = sm.fit_resample(features, labels)
    
    return features, labels

def _normalize(array):
    """
    Zero mean unit variance array normalization

    Args:
        array : numpy array
    Returns:
        norm : numpy array
            array with normalized values
    """
    norm = array - np.mean(array)
    norm = norm / np.std(array)
    return norm